//
// meli
//
// Copyright 2023 Manos Pitsidianakis
//
// This file is part of meli.
//
// meli is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// meli is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with meli. If not, see <http://www.gnu.org/licenses/>.

//use super::ViewFilter;
//use crate::melib::{Attachment, AttachmentBuilder};
//use crate::Context;

#[test]
fn test_view_filter_text_plain() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "text/plain");
}

#[test]
fn test_view_filter_text_html() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "text/html");
}

#[test]
fn test_view_filter_multipart_alternative_plain_and_html() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "text/plain");
}

#[test]
fn test_view_filter_multipart_alternative_empty_plain_and_html() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "text/html");
}

#[test]
fn test_view_filter_multipart_digest() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "multipart/digest");
}

#[test]
fn test_view_filter_multipart_mixed() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "multipart/mixed");
}

#[test]
fn test_view_filter_multipart_related() {
    println!("[ref:TODO]");
    //let bytes = b"";
    //let tempdir = tempfile::tempdir().unwrap();
    //let mut ctx = Context::new_mock(&tempdir);
    //let att: Attachment = AttachmentBuilder::new(bytes).build();
    //let value = ViewFilter::new_attachment(&att, &mut ctx).unwrap();
    //assert_eq!(&value.content_type.to_string(), "text/related");
}
